{
  "type": "record",
  "Aspect": {
    "name": "partitionsSummary"
  },
  "name": "PartitionsSummary",
  "namespace": "com.linkedin.pegasus2avro.dataset",
  "fields": [
    {
      "type": [
        "null",
        {
          "type": "record",
          "name": "PartitionSummary",
          "namespace": "com.linkedin.pegasus2avro.dataset",
          "fields": [
            {
              "type": "string",
              "name": "partition",
              "doc": "A unique id / value for the partition for which statistics were collected,\ngenerated by applying the key definition to a given row."
            },
            {
              "type": [
                "null",
                "long"
              ],
              "name": "createdTime",
              "default": null,
              "doc": "The created time for a given partition."
            },
            {
              "type": [
                "null",
                "long"
              ],
              "name": "lastModifiedTime",
              "default": null,
              "doc": "The last modified / touched time for a given partition."
            }
          ],
          "doc": "Defines how the data is partitioned"
        }
      ],
      "name": "minPartition",
      "default": null,
      "doc": "The minimum partition as ordered"
    },
    {
      "type": [
        "null",
        "com.linkedin.pegasus2avro.dataset.PartitionSummary"
      ],
      "name": "maxPartition",
      "default": null,
      "doc": "The maximum partition as ordered"
    }
  ],
  "doc": "Defines how the data is partitioned for Data Lake tables (e.g. Hive, S3, Iceberg, Delta, Hudi, etc)."
}