# mypy: ignore-errors
# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# pylint: skip-file
# fmt: off
# isort: skip_file
from .....schema_classes import AccessClass
from .....schema_classes import AccessLevelClass
from .....schema_classes import AuditStampClass
from .....schema_classes import BrowsePathEntryClass
from .....schema_classes import BrowsePathsClass
from .....schema_classes import BrowsePathsV2Class
from .....schema_classes import ChangeAuditStampsClass
from .....schema_classes import CostClass
from .....schema_classes import CostCostClass
from .....schema_classes import CostCostDiscriminatorClass
from .....schema_classes import CostTypeClass
from .....schema_classes import DataPlatformInstanceClass
from .....schema_classes import DeprecationClass
from .....schema_classes import DocumentationClass
from .....schema_classes import DocumentationAssociationClass
from .....schema_classes import EdgeClass
from .....schema_classes import EmbedClass
from .....schema_classes import FabricTypeClass
from .....schema_classes import FieldFormPromptAssociationClass
from .....schema_classes import FormAssociationClass
from .....schema_classes import FormPromptAssociationClass
from .....schema_classes import FormPromptFieldAssociationsClass
from .....schema_classes import FormVerificationAssociationClass
from .....schema_classes import FormsClass
from .....schema_classes import GlobalTagsClass
from .....schema_classes import GlossaryTermAssociationClass
from .....schema_classes import GlossaryTermsClass
from .....schema_classes import IncidentSummaryDetailsClass
from .....schema_classes import IncidentsSummaryClass
from .....schema_classes import InputFieldClass
from .....schema_classes import InputFieldsClass
from .....schema_classes import InstitutionalMemoryClass
from .....schema_classes import InstitutionalMemoryMetadataClass
from .....schema_classes import MLFeatureDataTypeClass
from .....schema_classes import MediaClass
from .....schema_classes import MediaTypeClass
from .....schema_classes import MetadataAttributionClass
from .....schema_classes import OperationClass
from .....schema_classes import OperationSourceTypeClass
from .....schema_classes import OperationTypeClass
from .....schema_classes import OriginClass
from .....schema_classes import OriginTypeClass
from .....schema_classes import OwnerClass
from .....schema_classes import OwnershipClass
from .....schema_classes import OwnershipSourceClass
from .....schema_classes import OwnershipSourceTypeClass
from .....schema_classes import OwnershipTypeClass
from .....schema_classes import RoleAssociationClass
from .....schema_classes import SiblingsClass
from .....schema_classes import StatusClass
from .....schema_classes import SubTypesClass
from .....schema_classes import TagAssociationClass
from .....schema_classes import TimeStampClass
from .....schema_classes import VersionTagClass
from .....schema_classes import WindowDurationClass


Access = AccessClass
AccessLevel = AccessLevelClass
AuditStamp = AuditStampClass
BrowsePathEntry = BrowsePathEntryClass
BrowsePaths = BrowsePathsClass
BrowsePathsV2 = BrowsePathsV2Class
ChangeAuditStamps = ChangeAuditStampsClass
Cost = CostClass
CostCost = CostCostClass
CostCostDiscriminator = CostCostDiscriminatorClass
CostType = CostTypeClass
DataPlatformInstance = DataPlatformInstanceClass
Deprecation = DeprecationClass
Documentation = DocumentationClass
DocumentationAssociation = DocumentationAssociationClass
Edge = EdgeClass
Embed = EmbedClass
FabricType = FabricTypeClass
FieldFormPromptAssociation = FieldFormPromptAssociationClass
FormAssociation = FormAssociationClass
FormPromptAssociation = FormPromptAssociationClass
FormPromptFieldAssociations = FormPromptFieldAssociationsClass
FormVerificationAssociation = FormVerificationAssociationClass
Forms = FormsClass
GlobalTags = GlobalTagsClass
GlossaryTermAssociation = GlossaryTermAssociationClass
GlossaryTerms = GlossaryTermsClass
IncidentSummaryDetails = IncidentSummaryDetailsClass
IncidentsSummary = IncidentsSummaryClass
InputField = InputFieldClass
InputFields = InputFieldsClass
InstitutionalMemory = InstitutionalMemoryClass
InstitutionalMemoryMetadata = InstitutionalMemoryMetadataClass
MLFeatureDataType = MLFeatureDataTypeClass
Media = MediaClass
MediaType = MediaTypeClass
MetadataAttribution = MetadataAttributionClass
Operation = OperationClass
OperationSourceType = OperationSourceTypeClass
OperationType = OperationTypeClass
Origin = OriginClass
OriginType = OriginTypeClass
Owner = OwnerClass
Ownership = OwnershipClass
OwnershipSource = OwnershipSourceClass
OwnershipSourceType = OwnershipSourceTypeClass
OwnershipType = OwnershipTypeClass
RoleAssociation = RoleAssociationClass
Siblings = SiblingsClass
Status = StatusClass
SubTypes = SubTypesClass
TagAssociation = TagAssociationClass
TimeStamp = TimeStampClass
VersionTag = VersionTagClass
WindowDuration = WindowDurationClass

# fmt: on
